type coin = int

module OrderedCoin =
  struct
    type t = coin
    let compare = Pervasives.compare
  end
      
module CoinSet = Set.Make(OrderedCoin)

type coinSet = CoinSet.t

type weighing = coinSet*coinSet

(* gnre tous les ensembles de n pices < max_coin *)
let generate_all_sets card max_coin =
  
  (* cre une liste de pice partant de i jusqu' max inclus *)
  let rec growing_coins i max =
    if i>max
    then []
    else i::(growing_coins (i+1) max)
  in

  (* premier et dernier lment gnr *)
  let first_elt = growing_coins 1 card
  and last_elt = growing_coins (max_coin-card+1) max_coin in

  (* construit l'lment suivant elt en augmentant la pice la plus *)
  (* "profonde", aucune pice ne peut dpasser max *)
  let rec built_next_elt elt max =
    match (List.rev elt) with
    | hd::tail ->
	if hd<max
	(* cas o la dernire pice peut tre augmente *)
	then List.rev ((hd+1)::tail)
	(* cas o il faut augmenter une pice autre que la dernire *)    
	else
	  let new_beginning = built_next_elt (List.rev tail) (max-1) in
	  new_beginning@[(List.hd (List.rev new_beginning))+1]
    | _ -> failwith "built_next_elt: lment vide"
  in

  (* transforme une liste en ensemble *)
  let built_set coin_list = List.fold_right CoinSet.add coin_list CoinSet.empty in

  (* gnre un lment et passe au suivant si ce n'est pas le dernier *)
  let rec generate elt =  
    if List.for_all2 (=) elt last_elt
    then [built_set elt]
    else (built_set elt)::(generate (built_next_elt elt max_coin))
  in

  generate first_elt;;


(* gnre toutes les peses possibles entre ensemble de taille card avec *)
(* des pices infrieures ou gales  max_coin *)
let generate_weighings card max_coin =

  (* liste des ensembles  card pices infrieures ou gales  max_coin *)
  let all_sets = generate_all_sets card max_coin in

  (* gnre la srie pese qui compare le premier lment d'une liste  *)
  (* tous les autres *)     
  let generate_weighings_serie = function
    | [] -> []
    | _::[] -> []
    | first::tail ->
	(* deux ensembles sont comparables s'ils n'ont aucune pice en commun *)
	let are_comparable s1 s2 =
	  CoinSet.equal (CoinSet.inter s1 s2) CoinSet.empty in
	(* liste des ensembles  comparer au premier lment *)
	let to_compare = List.filter (are_comparable first) tail in
	let build_weighing s1 s2 = (s1,s2) in
	List.map (build_weighing first) to_compare
  in

  (* gnre toutes les sries de peses pour tous les ensembles *)
  (*  chaque fois on ne pse un ensemble qu'avec les ensembles qui le suive *)
  (* dans la liste des ensembles, on vite ainsi de faire les peses 1-2 *)
  (* et 2-1 *)
  let rec generate_series sets =
    if (List.length sets)<2
    then []
    else (generate_weighings_serie sets)@(generate_series (List.tl sets))
  in

  generate_series all_sets;;
	

(* gnre toutes les peses entre ensembles de taille <= card *)
let generate_all_weighings card max_coin =
  let result = ref [] in
  for i = 1 to card do
    result := !result@(generate_weighings i max_coin)
  done;
  !result;;

    
(* vrai si la pice est dans l'ensemble *)
let in_set = CoinSet.mem;;

(* renvoie la liste des lments d'un ensemble *)
let elements = CoinSet.elements;;

(* renvoie l'ensemble vide *)
let empty_set = CoinSet.empty;;

(* renvoie une pese par dfaut *)
let default_weighing = (CoinSet.empty,CoinSet.empty);;

(* affiche une pese *)    
let print_weighing w =
  (* pices  gauche et  droite *)
  let left = fst w and right = snd w in
  let l_coin_arr = Array.of_list (CoinSet.elements left)
  and r_coin_arr = Array.of_list (CoinSet.elements right) in
  let n = Array.length l_coin_arr in
  for i = 0 to n-1 do
    print_int l_coin_arr.(i);
    print_string " ";
  done;
  print_string "| ";
  for i = 0 to n-1 do
    print_int r_coin_arr.(i);
    print_string " ";
  done;;
